import xlrd
import logging
import sys
import json

    
def read_excel_workbook(file_path):
    book = xlrd.open_workbook(file_path)
    logging.debug("The number of worksheets is {0}".format(book.nsheets))
    logging.debug("Worksheet name(s): {0}".format(book.sheet_names()))
    workbook_values = []
    num_sheets = book.nsheets
    if book.nsheets == 0:
        logging.error("No Sheets present in the file")
        return None
    for sheet_idx in range(0, num_sheets):
        min_row_id = 999999
        max_row_id = 0
        min_col_id = 999999
        max_col_id = 0

        sh = book.sheet_by_index(sheet_idx)
        for row_id in range(0, sh.nrows):
            for col_id in range(0, sh.ncols):
                if sh.cell_value(row_id, col_id) != xlrd.empty_cell.value:
                    if max_col_id < col_id:
                        max_col_id = col_id
                    if min_col_id > col_id:
                        min_col_id = col_id
                    if max_row_id < row_id:
                        max_row_id = row_id
                    if min_row_id > row_id:
                        min_row_id = row_id
        sheet_matrix = []
        for row_id in range(min_row_id, max_row_id + 1):
            row_data = sh.row_slice(row_id, min_col_id, max_col_id + 1)
            row_values = []
            for data in row_data:
                if data.value == xlrd.empty_cell.value:
                    row_values.append("")
                else:
                    row_values.append(data.value)
            sheet_matrix.append(row_values)
        workbook_values.append(sheet_matrix)
    return json.dumps(workbook_values)


def log_helper():
    root = logging.getLogger()
    root.setLevel(logging.DEBUG)
    ch = logging.StreamHandler(sys.stdout)
    ch.setLevel(logging.DEBUG)
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    ch.setFormatter(formatter)
    root.addHandler(ch)

if __name__ == "__main__":
    log_helper()
    print((read_excel_workbook('./tests/Diarrhoea.xlsx')))



import logging
import psycopg2
import socket


class Db(object):
    params = None
    connection = None
    logger = None

    def __init__(self, params):
        self.params = params
        self.params['keepalives'] = 1
        self.params['keepalives_idle'] = 120
        self.params['keepalives_interval'] = 120
        self._make_connection()

    def _make_connection(self):
        # logging.debug(self.params) exposing connection info in debug
        try:
            self.connection = psycopg2.connect(**self.params)
        except psycopg2.DatabaseError as e:
            logging.error("Could not connect to the server: %s" % e)
        except psycopg2.Error as e:
            logging.error("Error %s" % e)
        except Exception as e:
            logging.error("Error %s" % e)

    def get_connection(self):
        return self.connection


    '''
    Cursor will be created per crawler
    '''
    def get_cursor(self):
        try:
            cursor = self.connection.cursor()
            return cursor
        except Exception as err:
            logging.warning("Connection seems to have expired, remaking it")
            try:
                self._make_connection()
                cursor = self.connection.cursor()
                return cursor
            except Exception as e:
                logging.error("Connection could not be made: %s" % e)
                return None

    def close_cursor(self, cursor):
        logging.debug("Closing cursor...")
        cursor.close()

    def commit(self):
        self.connection.commit()

    def close_connection(self):
        logging.info("Closing connection...")
        self.connection.close()



"""
Program's aim is to get tables from a given html
Pure functional.
"""
import urllib.parse
from bs4 import BeautifulSoup

def _get_soup(html, elem):
    """
    Soup html element
    :param html:
    :param elem:
    :return:
    """
    soup = BeautifulSoup(html)
    tables = soup.findAll(elem)
    return tables


def _get_fixed_soup(html):
    """
    Fix the soup
    :param html:
    :return:
    """
    tables = _get_soup(html, 'table')
    if tables is not None and len(tables) > 0:
        return tables
    # Sometime we get just tr td, in a small html snippet.
    trs = _get_soup(html, 'tr')
    if trs is not None:
        return _get_soup("<table>" + html + "</table>", 'table')
    return None


def _get_header_cells(soup):
    all_th = []
    th = soup('th')
    for i in th:
        all_th.append(i.text)
    return all_th


def _get_caption_cells(soup):
    all_th = []
    th = soup('caption')
    for i in th:
        all_th.append(i.text)
    return all_th


def get_table(html):
    """
    :param html: html of a page.
    :return: (number of tables found, {'header' : top_row, 'data' : other_rows})
    """
    all_tables = []
    soup = _get_fixed_soup(html)
    if not soup:
        return 0, None

    count = 0
    if type(soup) == list:
        for s in soup:
            rows = [[cell.text for cell in row("td")]
                         for row in s("tr")]
            count = count + 1
            first_row = _get_header_cells(s)
            caption = _get_caption_cells(s)
            all_tables.append({'header' : first_row, 'data': rows, 'title': caption})
    else:
        rows = [[cell.text for cell in row("td")]
                         for row in soup("tr")]
        count = count + 1
        first_row = _get_header_cells(soup)
        caption = _get_caption_cells(soup)
        all_tables.append({'header' : first_row, 'data': rows, 'title': caption})
    return count, all_tables


def get_table_with_links(html, base_url):
    """
    Fetch all tables in the html that contain links,
    returns a list of dictionaries with {header : [], data : [[]], num_links : 4}
    :param html:
    :return:
    """
    all_tables = []
    soup = _get_fixed_soup(html)
    if not soup:
        return 0, None

    count = 0
    if type(soup) == list:
        for s in soup:
            rows = [[cell.text for cell in row("td")]
                    for row in s("tr")]
            links = [[_fix_url(cell, base_url) for cell in row("td")]
                    for row in s("tr")]
            count = count + 1
            first_row = _get_header_cells(s)
            caption = _get_caption_cells(s)
            all_tables.append({'header': first_row, 'data': rows, 'title': caption, 'links': links})
    else:
        rows = [[cell.text for cell in row("td")]
                for row in soup("tr")]

        links = [[cell.findAll('a') for cell in row("td")]
                 for row in soup("tr")]
        count = count + 1
        first_row = _get_header_cells(soup)
        caption = _get_caption_cells(soup)
        all_tables.append({'header': first_row, 'data': rows, 'title': caption, 'links' : links})
    return count, all_tables


def _fix_url(soup, base_url):
    # extract url from url_raw
    aa = soup.findAll("a")
    a_href = None
    for a in aa:
        try:
            a_href_rel = a['href']
            a_href = urllib.parse.urljoin(base_url, a_href_rel)
        except Exception as e:
            continue
    return a_href

if __name__ == "__main__":
    import requests
    base_url = "http://www.jreda.com/tenders/tenders.htm"
    r = requests.get(base_url)
    num, tables = get_table_with_links(r.content, base_url)
    print(num)
    for table in tables:
        for link in table["links"]:
            print(link)


from datetime import datetime, timedelta
import pytz
import time
import calendar

def _utc_to_local(utc_dt, tz):
    local_dt = utc_dt.replace(tzinfo=pytz.utc).astimezone(tz)
    return tz.normalize(local_dt)

def get_ist_date():
    india = pytz.timezone('Asia/Kolkata')
    x = _utc_to_local(datetime.utcnow(), india)
    return x.strftime('%Y-%m-%d')

def get_ist_time():
    india = pytz.timezone('Asia/Kolkata')
    x = _utc_to_local(datetime.utcnow(), india)
    return x.strftime('%H:%M:%S')

def get_ist_datetime():
    india = pytz.timezone('Asia/Kolkata')
    x = _utc_to_local(datetime.utcnow(), india)
    return x.strftime('%Y%m%d%H%M%S')


def get_hawker_date(day_gap):
    india = pytz.timezone('Asia/Kolkata')
    x = _utc_to_local(datetime.utcnow(), india)
    if day_gap != 0:
       x = _utc_to_local(datetime.utcnow() -  timedelta(days=day_gap), india)
    return x.strftime('%Y%m%d')

def get_epoch_date(hawker_time):
    t = time.strptime(hawker_time, '%Y%m%d%H%M%S')
    return int(calendar.timegm(t)/60)


"""
This file contains the code for making chunks out of a line or a phrase.
"""


def acceptable_word(word):
    from spacy.lang.en.stop_words import STOP_WORDS
    """Checks conditions for acceptable word: length, stopword."""
    accepted = bool(2 <= len(word) <= 40 and word.lower() not in STOP_WORDS)
    return accepted


def get_chunks(text):
    import spacy
    nlp = spacy.load('en')
    doc = nlp(text)
    chunks = []
    for np in doc.noun_chunks:
        chunk = np.text
        chunk = ' '.join([word for word in chunk.split() if acceptable_word(word)])
        chunks.append(chunk)
    return chunks


if __name__ == "__main__":
    text = """Ar and mo various roads under sub divn m 2112 dg 2016 17 sh provision of maintenance van with required labour and t and p"""
    print((get_chunks(text)))
    text = """Construction of police station, fire station and rajkishore vidyapitha at arisol, infovalley,khurda """
    print((get_chunks(text)))
    text = """Timber treatment impregnation plant """
    print((get_chunks(text)))
    text = """Timber treatment impregnation plant """
    print((get_chunks(text)))



import logging


def make_postgres_delete_statement(table, kv_map, debug):
    _prefix = "DELETE FROM "
    keys = " and ".join([k + "=%s" for k in list(kv_map.keys())])
    statement = " ".join([_prefix, table, " where ", keys])
    if debug:
        logging.warning("Deleting from Db: %s, %s" % (statement, list(kv_map.values())))
    return statement, list(kv_map.values())



from . import pg_python
import unittest
import logging
import requests

COL_1 = "col1"
COL_2 = "col2"
COL_3 = "col3"
COL_4 = "col4"
COL_5 = "int_value"
UPDATE = "update"
test_table = "pg_python_test"


class TestTests(unittest.TestCase):
    """
    Please run the following in your local database before running the test cases:
    1. A user postgres is required to be present with password "@pgtest"
    2. create database test_db;
    3. create table pg_python_test (col1 text, col2 text, col3 int, col4 text, int_value int)
    """
    def setUp(self):
        pg_python.pg_server("test_db", "postgres", "@pgtest", "localhost", False)

    def test_read_raw(self):
        create_rows()
        title1 = pg_python.read_raw("select col1,col3 from "+test_table+" where col2=%s",('read',))
        self.assertEqual(title1[0][0],"title1")
        self.assertEqual(title1[0][1], 76)
        print("Read raw done")
        clear_table()

    def test_write_raw(self):
        create_rows()
        values = ['title7','read7',77]
        pg_python.write_raw("insert into "+test_table+" values (%s,%s,%s)",values)
        title1 = pg_python.read(test_table,[COL_2,COL_3],{COL_1:'title7'})
        self.assertEqual(title1[0][COL_2], "read7")
        self.assertEqual(title1[0][COL_3], "77")
        print("Write raw done")
        clear_table()

    def test_update(self):
        create_rows()
        pg_python.update(test_table,{COL_4:'updated_name1'},{COL_1:'title1%'},clause='ilike')
        title1 = pg_python.read(test_table,[COL_4],{COL_1:'title15'})
        self.assertEqual(title1[0][COL_4],'updated_name1')
        print("Update single row done")
        clear_table()

    def test_multiple_update(self):
        create_rows()
        dict_lst =[
            {COL_1:'title1', UPDATE:'updated_name1'},
            {COL_1: 'title2', UPDATE: "update'd_name2"}
        ]
        pg_python.update_multiple(test_table,COL_4,[COL_1],dict_lst)
        title1 = pg_python.read(test_table,[COL_4],{COL_1:'title1'})
        title2 = pg_python.read(test_table, [COL_4], {COL_1: 'title2'})
        self.assertEqual(title1[0][COL_4],'updated_name1')
        self.assertEqual(title2[0][COL_4], "update'd_name2")
        print("Update multiple done")
        clear_table()

    def test_multiple_insert(self):
        column_to_insert = [COL_1, COL_3]
        insert_dict_list = [
            {COL_1: "insert1", COL_3: 1},
            {COL_3: 2, COL_1: "insert2"},
            {COL_1: "insert3", COL_3: 3}
        ]
        pg_python.insert_multiple(test_table,column_to_insert,insert_dict_list)
        val1 = pg_python.read(test_table,[COL_1],{COL_3:1})
        val2 = pg_python.read(test_table, [COL_1], {COL_3: 2})
        val3 = pg_python.read(test_table, [COL_1], {COL_3: 3})
        values = [val1[0][COL_1],val2[0][COL_1],val3[0][COL_1] ]
        self.assertEqual(values,["insert1","insert2","insert3"])
        print("Insert multiple done")
        clear_table()

    def test_update_raw(self):
        create_rows()
        updates = pg_python.update_raw("update "+ test_table + " set " + COL_4 + "= 'updated_name1' where " + COL_1 + " ilike 'title1%'" )
        title1 = pg_python.read(test_table, [COL_4], {COL_1: 'title15'})
        self.assertEqual(title1[0][COL_4], 'updated_name1')
        self.assertEqual(updates, 2)
        print("Update raw done")
        clear_table()

    def test_read_in(self):
        create_rows()
        values = ['title15','title1','title2','title3']
        rows = pg_python.read(test_table, [COL_1], {COL_1: values}, clause=" in ")
        read_values = []
        for row in rows:
            read_values.append(row.get(COL_1))
        self.assertEqual(sorted(read_values), sorted(values))
        print("Read in done")
        clear_table()

    def test_read_null(self):
        create_rows()
        values = ['null']
        rows = pg_python.read(test_table, [COL_1,COL_4], {COL_1: 'title-null', COL_2: values[0]})
        read_values = []
        read_values_4 = []
        for row in rows:
            read_values.append(row.get(COL_1))
            read_values_4.append(row.get(COL_4))
        self.assertEqual(sorted(read_values), ['title-null'])
        self.assertEqual(sorted(read_values_4), ['sec-value'])
        print("Read in done")
        clear_table()


    def test_read_simple(self):
        create_rows()
        rows = pg_python.read(test_table, [COL_1], {COL_2: 'read'})
        self.assertEqual(len(rows), 1)
        self.assertEqual(rows[0][COL_1], 'title1')
        print("Read simple done")
        clear_table()

    def test_read_greater_than(self):
        create_rows()
        rows = pg_python.read(test_table, [COL_1], {COL_5 + " >": 10})
        self.assertEqual(len(rows), 1)
        self.assertEqual(rows[0][COL_1], 'title6')
        print("Read greater than done")
        clear_table()

    def test_delete(self):
        create_rows()
        where_kv_map={COL_1:'title15'}
        pg_python.delete(test_table, where_kv_map)
        read=pg_python.read(test_table,[COL_2], {COL_1:'title15'})
        self.assertEqual(read,[])
        print("Delete row done")
        clear_table()

class UpdateTests(unittest.TestCase):
    def setUp(self):
        pg_python.pg_server("test_db", "postgres", "@pgtest", "localhost", False)

    def test_update_multicol(self):
        create_rows()
        cols_to_update = [COL_2, COL_3]
        dict_lst =[
            {'where':{COL_1:'title1'}, UPDATE:{COL_2 : 'updated_col2a', COL_3:33}},
             {'where':{COL_1:'title2'}, UPDATE:{COL_2 : 'updated_col2b', COL_3:34}
            }
        ]
        pg_python.update_multiple_col(test_table,cols_to_update,[COL_1],dict_lst)
        title1 = pg_python.read(test_table,[COL_2, COL_3],{COL_1:'title1'})
        title2 = pg_python.read(test_table, [COL_2, COL_3], {COL_1: 'title2'})
        self.assertEqual(title1[0][COL_2],'updated_col2a')
        self.assertEqual(title1[0][COL_3], '33')
        self.assertEqual(title2[0][COL_2], 'updated_col2b')
        self.assertEqual(title2[0][COL_3], '34')
        print("Update multiple row multiple column done")
        clear_table()

    def test_update_multicol_2testcond(self):
        create_rows()
        cols_to_query = [COL_1, COL_3]
        cols_to_update = [COL_2]
        dict_lst =[
            {'where':{COL_1:'title1', COL_3:76}, UPDATE:{COL_2 : 'updated_col2a'}}
        ]
        num = pg_python.update_multiple_col(test_table,cols_to_update,cols_to_query,dict_lst)
        self.assertEqual(num['updated_records'], 1)
        title1 = pg_python.read(test_table,[COL_2],{COL_1:'title1'})

        self.assertEqual(title1[0][COL_2],'updated_col2a')
        print("Update multiple row multiple column with two test conditions done")
        clear_table()

def create_rows():
    pg_python.write(test_table, {COL_1: "title1", COL_2: "read", COL_3: 76, COL_4: "reeer"})
    pg_python.write(test_table, {COL_1: "title2", COL_2: "read2", COL_3: 77, COL_4: "reeer"})
    pg_python.write(test_table, {COL_1: "title3", COL_2: "read3", COL_3: 77, COL_4: "reeer"})
    pg_python.write(test_table, {COL_1: "title4", COL_2: "read4", COL_3: 77, COL_4: "reeer"})
    pg_python.write(test_table, {COL_1: "title15", COL_2: "read5", COL_3: 77, COL_4: "reeer"})
    pg_python.write(test_table, {COL_1: "title6", COL_2: "read6", COL_3: 77, COL_4: "reeer", COL_5: 20})
    pg_python.write(test_table, {COL_1: "title-null", COL_4: "sec-value"})

def clear_table():
    pg_python.write_raw("Delete from %s"%(test_table), None)



import logging

def make_postgres_write_statement(table, kv_map, debug=True):
    _prefix = "INSERT INTO"
    keys = ",".join(list(kv_map.keys()))
    values = []
    for value in list(kv_map.values()):
      if type(value) == bool:
        if value == True:
          values.append('true')
        else:
          values.append('false')
      else:
        values.append(value)
  
    value_proxy_array = ["%s"] * len(kv_map)
    value_proxy_string = ", ".join(value_proxy_array)
    statement = " ".join([_prefix, table, "(", keys ,")", "VALUES", "(", value_proxy_string ,")"])
    if debug:
        logging.info("Writing into Db: %s, %s" % (statement, values))
    return statement, list(kv_map.values())



def get_multi_insert_str(columns_to_insert,insert_values_dict_lst):
    """
    get_multi_insert_str creates the placeholder string for multiple insertion values.
    :param columns_to_insert:
    :param insert_values_dict_lst:
    :return:
    """
    dict_lst= []
    placeholder_str = ["%s"]*len(columns_to_insert)
    row_str = ",".join(placeholder_str)
    row_str = "("+row_str+")"
    row = [row_str]
    multi_row_str_lst = row*len(insert_values_dict_lst)
    multi_row_str = ",".join(multi_row_str_lst)
    return multi_row_str

def get_multi_values(columns_to_insert,insert_values_dict_lst):
    """
    returns the values for the placeholders in query.
    :param columns_to_insert:
    :param insert_values_dict_lst:
    :return:
    """
    values = []
    for value_dict in insert_values_dict_lst:
        for col in columns_to_insert:
            value = value_dict[col]
            values.append(value)
    return  values


def make_postgres_write_multiple_statement(table, columns_to_insert_lst, insert_values_dict_lst, print_debug_log=True):
    """
    make_postgres_write_multiple_statement generates the posgresql query.

    :param table:
    :param columns_to_insert_lst:
    :param insert_values_dict_lst:
    :param print_debug_log:
    :return:
    """
    prefix = "INSERT INTO"
    TABLE_NAME = table
    columns_str = "("+",".join(columns_to_insert_lst) + ")"
    values_placeholder = get_multi_insert_str(columns_to_insert_lst, insert_values_dict_lst)
    values = get_multi_values(columns_to_insert_lst,insert_values_dict_lst)
    statement = " ".join([prefix, TABLE_NAME, columns_str, "VALUES", values_placeholder])
    return statement,values




import logging


def make_postgres_read_statement(table, kv_map, keys_to_get, limit, order_by,
                                 order_type, debug, clause, group_by, join_clause):
    _prefix = "SELECT"
    _join_by = " " + join_clause + " "
    _table_string = " ".join(["FROM", table])
    null_keys = []

    for key in kv_map.keys():
        kval = kv_map.get(key)
        try:
            if str(kval).lower() == 'null':
                null_keys.append(key)
        except Exception as e:
            continue

    for key in null_keys:
        if key in kv_map:
            kv_map.pop(key, None)

    _key_string = _join_by.join([k + clause + "%s" for k in list(kv_map.keys())])
    _null_key_string = ""
    if len(null_keys) > 0:
        _null_key_string = _join_by.join([k + " is null " for k in null_keys])
    if len(_null_key_string) > 2:
        if len(_key_string) > 2:
            _key_string = _key_string + " and " + _null_key_string
        else:
            _key_string = _key_string + " " + _null_key_string

    values = list(kv_map.values())

    if clause.strip().lower() == "in":
        values = []
        _key_string = _join_by.join([k + clause + "(" + ",".join("'" + x + "'" for x in kv_map[k]) + ")" for k in list(kv_map.keys())])

    statement = " ".join([_prefix, ", ".join(sorted(keys_to_get)), _table_string])
    if len(list(kv_map.keys())) > 0 or len(null_keys) > 0:
      statement = " ".join([_prefix, ", ".join(sorted(keys_to_get)), _table_string, "WHERE", _key_string])
    if group_by is not None:
        statement += " GROUP BY " + group_by
    if order_by is not None:
        statement += " ORDER BY " + order_by + " " + order_type
    if limit is not None:
        statement += " LIMIT " + str(limit)
    if debug:
        logging.info("Reading From Db: %s, %s" %(statement, list(kv_map.values())))
    return statement, values


def prepare_values(all_values, keys_to_get, cols_keep_raw_type=[]):
    ret_val = []
    if all_values is None:
        return None
    k = sorted(keys_to_get)
    for row in all_values:
        row_kv = {}
        if len(row) == len(keys_to_get):
            for idx in range(0, len(row)):
                col_name = k[idx]
                if col_name in cols_keep_raw_type:
                    row_kv[k[idx]] = row[idx]
                else:
                    row_kv[k[idx]] = str(row[idx])
        else:
            logging.error("Number of keys to be fetched are not correct")
            continue
        ret_val.append(row_kv)
    return ret_val




from selenium.webdriver import Chrome
from scrapy.http.response.html import HtmlResponse
import logging

"""
Gathering the HTML response from chrome.
retries is equal to 3
"""

MAX_NUM_RETRY = 3


def get_response(url):
    num_retries = 0
    driver = None
    while num_retries < MAX_NUM_RETRY:
        try:
            driver = Chrome()
            driver.get(url)
            response = HtmlResponse(url=url, body=driver.page_source, encoding='utf-8')
            driver.close()
            return response
        except Exception as e:
            logging.error("Exception %s" % e)
            num_retries += 1
    logging.error("Could not fetch the url")
    if driver is not None:
        driver.close()


if __name__ == "__main__":
    response = get_response("https://google.com")
    print(response)


from ._db_object import Db
from ._write import make_postgres_write_statement,make_postgres_write_multiple_statement
from ._read import make_postgres_read_statement, prepare_values
from ._update import make_postgres_update_statement
from ._update import make_postgres_update_multiple_statement
from ._delete import make_postgres_delete_statement


local_db = None

def _get_db():
    global local_db
    return local_db

print_debug_log = True
params_map = {}


def pg_server(db_name, username, password, host_address, debug=True, send_keep_alive_probes=False, socket_idle_time=120):
  global local_db, print_debug_log, params_map
  params_map = {
    'dbname': db_name,
    'user': username,
    'password': password,
    'host': host_address,
    'send_keep_alive_probes': send_keep_alive_probes,
    'socket_idle_time': socket_idle_time,
    }
  local_db = Db(params_map)
  print_debug_log = debug
  return local_db



def write(table, kv_map):
    """
    :param table: String.
    :param kv_map: Key values.
    :return success_bool:
    """
    global local_db, print_debug_log, params_map
    connection = local_db.get_connection()
    cursor = local_db.get_cursor()
    command, values = make_postgres_write_statement(table, kv_map, print_debug_log)
    try:
        cursor.execute(command, values)
        connection.commit()
    except Exception as e:
        print(("Db Cursor Write Error: %s" % e))
        local_db = Db(params_map)
        return False
    return True

def read(table, keys_to_get, kv_map, limit=None, order_by=None, order_type=None, clause = "=", group_by = None, join_clause = ' AND '):
    """
    :param table: String
    :param keys_to_get: list of strings
    :param kv_map: key value map, if this is None, then limit is maxed at 1000
    :param limit: None or integer
    :param order_by: None or must be of a type String
    :param order_type: String None, "ASC" or "DESC" only
    :return: values in an array of key value maps
    """
    error_return = None
    cursor = local_db.get_cursor()
    command, values = make_postgres_read_statement(table, kv_map, keys_to_get,
                                                   limit, order_by, order_type, print_debug_log,
                                                   clause, group_by, join_clause)
    try:
        cursor.execute(command, values)
        all_values = cursor.fetchall()
        return prepare_values(all_values, keys_to_get)
    except Exception as e:
        print(("Db Cursor Read Error: %s" % e))
        return []

def update(table, update_kv_map, where_kv_map, clause = '='):
    """
    :param table: table name, type string
    :param update_kv_map: the NEW keyvalue map for values to be updated
    :param where_kv_map: the kv map to search for values, all values ARE ANDed.
    :return: Success or Failure.
    """
    global local_db, print_debug_log, params_map
    connection = local_db.get_connection()
    cursor = local_db.get_cursor()
    command, values = make_postgres_update_statement(table, update_kv_map, where_kv_map,
                                                     clause, print_debug_log)
    try:
        cursor.execute(command, values)
        connection.commit()
    except Exception as e:
        print(("Db Cursor Update Error: %s" % e))
        local_db = Db(params_map)
        return False
    return True

def read_raw(command, values):
    """
    :param table: String
    :param keys_to_get: list of strings
    :param kv_map: key value map, if this is None, then limit is maxed at 1000
    :param limit: None or integer
    :param order_by: None or must be of a type String
    :param order_type: String None, "ASC" or "DESC" only
    :return: values in an array of key value maps
    """
    cursor = local_db.get_cursor()
    try:
        if values is not None:
          cursor.execute(command, values)
        else:
          cursor.execute(command)
        all_values = cursor.fetchall()
        return all_values
    except Exception as e:
        print(("Db Cursor Read Error: %s" % e))
        return []

def write_raw(command, values):
    """
    :params command, values. Execution commands dirctly for postgres
    """
    global local_db, print_debug_log, params_map
    connection = local_db.get_connection()
    cursor = local_db.get_cursor()
    try:
        cursor.execute(command, values)
        connection.commit()
    except Exception as e:
        print(("Db Cursor Write Error: %s" % e))
        local_db = Db(params_map)
        return False
    return True


def update_raw(command):
    """
    Update statement in the raw format,
    :param command: SQL command
    :return: number of rows affected
    """
    global local_db, print_debug_log, params_map
    connection = local_db.get_connection()
    cursor = local_db.get_cursor()
    try:
        cursor.execute(command)
        rowcount = cursor.rowcount
        connection.commit()
    except Exception as e:
        print(("Db Cursor Update Error: %s" % e))
        local_db = Db(params_map)
        return -1
    return rowcount


class bcolors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'

def print_ok(s):
    print((bcolors.OKGREEN + s + bcolors.ENDC))

def print_warn(s):
    print((bcolors.WARNING + s + bcolors.ENDC))

def print_fail(s):
    print((bcolors.FAIL + s + bcolors.ENDC))


def close():
    global local_db, print_debug_log, params_map
    connection = local_db.get_connection()
    cursor = local_db.get_cursor()
    local_db.close_cursor(cursor)
    local_db.close_connection()

def delete(table, where_kv_map):
    """
    Delete the rows resulting from the mentined kv map. No limit.
    :param table: table name, must be string
    :param where_kv_map: the kv map to search for values, all values ARE ANDed.
    :return: True or False
    """
    global local_db, print_debug_log, params_map
    connection = local_db.get_connection()
    cursor = local_db.get_cursor()
    command, values = make_postgres_delete_statement(table, where_kv_map, print_debug_log)
    try:
        cursor.execute(command, values)
        connection.commit()
    except Exception as e:
        print(("Db Cursor Delete Error: %s" % e))
        local_db = Db(params_map)
        return False
    return True

def check_parameters(column_to_update, columns_to_query_lst, query_values_dict_lst):
    """
    check_prarameters checks whether the passed parameters are valid or not.
    :param column_to_update: name of column that is to be updated.
    :param columns_to_query_lst: list of column names that is used in where clause.
    :param query_values_dict_lst: list of dictionaries containing values for where clause and target column.
    :return: boolean
    """
    # check if dimensions are correct.
    expected_length = 1 + len(columns_to_query_lst)
    all_columns_name = ["update"] + columns_to_query_lst
    flag =0
    for dict_val in query_values_dict_lst:
        # check dimensions.
        if len(dict_val)!= expected_length:
            print(("%s doesn't match the dimensions"%(dict_val)))
            return False

        # check columns present.
        for column in all_columns_name:
            if column not in dict_val:
                print(("%s column isn't present in dictionary"%(column)))
                return False
    return True


def update_multiple(table, column_to_update, columns_to_query_lst,
                    query_values_dict_lst):
    """
    Multiple update support in pg_python
    :param table: table to update into
    :param column_to_update: Single column for set clause
    :param columns_to_query_lst: column names for where clause
    :param query_values_dict_lst: values for where and Set.
    :return:
    """
    global local_db, print_debug_log, params_map
    connection = local_db.get_connection()
    cursor = local_db.get_cursor()
    is_pararmeters_correct = check_parameters(column_to_update, columns_to_query_lst, query_values_dict_lst)
    if not is_pararmeters_correct:
        print("ERROR in parameters passsed")
        return

    command,values= make_postgres_update_multiple_statement(table,
                                                              column_to_update,
                                                              columns_to_query_lst,
                                                              query_values_dict_lst,
                                                              print_debug_log)
    try:
        cursor.execute(command,values)
        connection.commit()
    except Exception as e:
        print(("Db Cursor update_multiple Error: %s" % e))
        local_db = Db(params_map)
        return False
    return True



def check_multiple_insert_param(columns_to_insert,insert_values_dict_lst):
    """
    Checks if the pararmeter passed are of correct order.
    :param columns_to_insert:
    :param insert_values_dict_lst:
    :return:
    """
    column_len = len(columns_to_insert)
    for row in insert_values_dict_lst:
        if column_len!= len(row):
            print(("%s doesn't match the dimensions" % (row)))
            return False
        for column in columns_to_insert:
            if column not in row:
                print(("%s column isn't present in dictionary"%(column)))
                return False
    return True



def insert_multiple(table, columns_to_insert_lst, insert_values_dict_lst):
    """
    Multiple row insert in pg_python
    :param table: table to insert into.
    :param columns_to_insert_lst: columns value provided.
    :param insert_values_dict_lst: values of corresponding columns.
    :return:
    """
    global local_db, print_debug_log, params_map
    connection = local_db.get_connection()
    cursor = local_db.get_cursor()
    is_pararmeters_correct = check_multiple_insert_param(columns_to_insert_lst, insert_values_dict_lst)
    if not is_pararmeters_correct:
        print("ERROR in parameters passsed")
        return
    command,values = make_postgres_write_multiple_statement(table, columns_to_insert_lst, insert_values_dict_lst, print_debug_log)
    try:
        cursor.execute(command, values)
        connection.commit()
    except Exception as e:
        print(("Db Cursor Write Error: %s" % e))
        local_db = Db(params_map)
        return False
    return True




import logging


def make_postgres_update_statement(table, kv_map, where_kv_map, clause, debug = True):
    _prefix = "UPDATE"
    clause = " " + clause + " "
    keys = ",".join([k + "=%s" for k in list(kv_map.keys())])
    where_keys = " AND ".join([k + clause + "%s" for k in list(where_kv_map.keys())])
    value_proxy_array = ["%s"] * len(kv_map)
    value_proxy_string = " , ".join(value_proxy_array)
    statement = " ".join([_prefix, table, "SET", keys, "WHERE", where_keys])
    if debug:
        logging.info("Updating into Db: %s, %s" %(statement, list(kv_map.values()) + list(where_kv_map.values())))
    return statement, list(kv_map.values()) + list(where_kv_map.values())


def get_from_clause(query_values_dict_lst,columns_to_query_lst):
    """
    returns from clause that contains tuples of placeholders
    :param query_values_dict_lst:
    :param columns_to_query_lst:
    :return:
    """
    from_str = "from (values "
    length = len(columns_to_query_lst)+1
    placeholder_str = ["%s"] * length
    row_str = ",".join(placeholder_str)
    row_str = "(" + row_str + ")"
    num_of_dict = len(query_values_dict_lst)
    multi_row_str = [row_str]*num_of_dict
    multi_row_str = ",".join(multi_row_str)
    from_str = from_str +multi_row_str +")"
    return from_str


def get_as_clause(columns_to_query_lst):
    """
    get_as_clause will return all column names tuples.
    :param columns_to_query_lst: columns for where clause
    :return:
    """
    column_str = ""
    for col in columns_to_query_lst:
        column_str = column_str + col + ","
    column_str += "update"
    as_clause = "as c(" + column_str + ")"
    return  as_clause


def get_where_clause(columns_to_query_lst):
    """
    get_where_clause returns the where clause from the given query list.
    :param columns_to_query_lst: columns for where clause.
    :return:
    """
    where_str = "where "
    equals_str =[]
    for row in columns_to_query_lst:
        temp_str =  "c." + row + " = t." + row
        equals_str.append(temp_str)
    joined_str = " AND ".join(equals_str)
    where_str = where_str + joined_str
    return where_str

def get_values(column_to_query_lst, query_values_dict_lst ):
    """
    makes flat list for update values.
    :param column_to_query_lst:
    :param query_values_dict_lst:
    :return:
    """
    column_to_query_lst.append("update")
    values = []
    for dict_row in query_values_dict_lst:
        for col in column_to_query_lst:
            values.append(dict_row[col])
    return values


def make_postgres_update_multiple_statement(table,column_to_update,
                                            columns_to_query_lst,
                                            query_values_dict_lst,
                                            print_debug_log = True,
                                            typecast_suffix = ""):
    """
    It makes query statement for updating multiple statements.
    :param table: table to update.
    :param column_to_update: column name that is to be updated.
    :param columns_to_query_lst: columns name that will we used for where clause.
    :param query_values_dict_lst: list of dictionary that contains values to update.:param print_debug_log:
    :param typecast_suffix:
    :return:
    """
    _prefix = "UPDATE"
    table_name = table + " as t"
    keys = column_to_update + " = c.update" + typecast_suffix + " "
    from_clause = get_from_clause(query_values_dict_lst, columns_to_query_lst)
    as_clause = get_as_clause(columns_to_query_lst)
    where_clause = get_where_clause(columns_to_query_lst)
    statement = " ".join([_prefix, table_name, "SET", keys, from_clause, as_clause, where_clause])
    values = get_values(columns_to_query_lst, query_values_dict_lst)

    if print_debug_log:
        logging.info("Updating multiple rows into db %s" % statement)

    return statement, values


def make_keys_multicol(columns_to_update_lst):
    """
        returns keys to be updated and new names of columns to be updated
        :param columns_to_update_lst:
        :return joined_str: part of postgres query with keys. E.g. "col1=c.updatecol1 , col2=c.updatecol2"
                update_lst: list of new column names in intermediate tables. E.g. [updatecol1, updatecol2]
    """
    key_equal = []
    update_lst = []
    for key in columns_to_update_lst:
        temp_str = key + " = c.update" + key
        update_lst.append("update" +key)
        key_equal.append(temp_str)
    joined_str = ",".join(key_equal)
    return joined_str, update_lst


def get_from_clause_multicol(query_values_dict_lst,columns_to_query_lst, columns_to_update_lst):
    """
        returns from clause that contains tuples of placeholders
        :param query_values_dict_lst:
        :param columns_to_update_lst:
        :param columns_to_query_lst:
        :return from_str:string. E.g "from (value (%s,%s),(%s,%s))"
    """
    from_str = "from (values "
    length = len(columns_to_query_lst)+len(columns_to_update_lst)
    placeholder_str = ["%s"] * length
    row_str = ",".join(placeholder_str)
    row_str = "(" + row_str + ")"
    num_of_dict = len(query_values_dict_lst)
    multi_row_str = [row_str]*num_of_dict
    multi_row_str = ",".join(multi_row_str)
    from_str = from_str +multi_row_str +")"
    return from_str


def get_as_clause_multicol(columns_to_query_lst, update_param_list):
    """
        get_as_clause will return tuple of column names of intermediate table c of postgresql query.
        :param columns_to_query_lst: columns for where clause.E.g [col1]
        :param update_param_list: new column names for columns to be updated.E.g [updatecol2,updatecol3]
        :return as_clause: string. E.g "as c(col1,updatecol2,updatecol3)"
        """
    column_str = []
    for col in columns_to_query_lst:
        column_str.append(col)
    for col in update_param_list:
        column_str.append(col)
    as_clause = ",".join(column_str)
    as_clause = "as c(" + as_clause +  ")"
    return as_clause


def get_values_multicol(columns_to_query_lst, columns_to_update_lst, query_values_dict_lst ):
    """
        makes flat list for update values.
        :param columns_to_query_lst:
        :param columns_to_update_lst:
        :param query_values_dict_lst:
        :return values: list of user-given values for multirow query with first params to
                query and then params to update. E.g ['col2query1','col2update1','col2query2','col2update2']
        """
    values = []
    for dict_row in query_values_dict_lst:
        for col in columns_to_query_lst:
            values.append(dict_row['where'][col])
        for col in columns_to_update_lst:
            values.append(dict_row['update'][col])
    return values


def check_parameters_multicol(columns_to_update_lst, columns_to_query_lst, query_values_dict_lst):
    """
    Checks parameters for multirow multicolumn update
    :param columns_to_update_lst: list of column names to be updated for set clause, list of strings
    :param columns_to_query_lst: list of column names to be queried for where clause, list of strings
    :param query_values_dict_lst: list of dictionaries with values for where query and update params. Has 2 items-
            update dictionary - with key='update' and value as dictionary of column name and values to be updated
            where dictionary - with key='where' and value as dictionary of column names and values to be queried
            e.g. [{'where':{'id':1}, 'update':{'col1':'ABC', 'col2':'XYZ'}}]
    :return: boolean
    """
    expected_length_query_cols = len(columns_to_query_lst)
    expected_length_target_cols = len(columns_to_update_lst)
    for dict_val in query_values_dict_lst:
        #Check length of query columns
        if len(dict_val['where'])!=expected_length_query_cols:
            logging.error("Expected %s fields in query list (where clause), instead received %s: %s" % (
            expected_length_query_cols, len(dict_val['where']), dict_val))
            return False
        #Check length of target columns
        if len(dict_val['update'])!=expected_length_target_cols:
            logging.error("Expected %s fields in update list, instead received %s: %s" % (
            expected_length_target_cols, len(dict_val['update']), dict_val))
            return False
        # check columns present in update params
        for key in dict_val['update']:
            if key not in columns_to_update_lst:
                logging.error(
                    "'%s' attribute to update isn't present in update list: %s" % (key, columns_to_update_lst))
                return False
        # check columns present in query where params
        for key in dict_val['where']:
            if key not in columns_to_query_lst:
                logging.error("'%s' attribute to query isn't present in query list: %s" % (key,columns_to_query_lst))
                return False
    return True

def make_postgres_update_multiple_column_statement(table, columns_to_update_lst, columns_to_query_lst,
                                                   query_values_dict_lst, print_debug_log = True):
    """
    Makes the postgresql statement for multiple column multiple row update
    :param table: table to update into, string
    :param columns_to_update_lst: list of column names to be updated for set clause, list of strings
    :param columns_to_query_lst: list of column names to be queried for where clause, list of strings
    :param query_values_dict_lst: list of dictionaries with values for where query and update params. Has 2 items-
            update dictionary - with key='update' and value as dictionary of column name and values to be updated
            where dictionary - with key='where' and value as dictionary of column names and values to be queried
            e.g. [{'where':{'id':1}, 'update':{'col1':'ABC', 'col2':'XYZ'}}]
    :return statement:final postgres query statement
            values: list of values given by user corresponding to placeholders in statement
    """
    _prefix = "UPDATE"
    table_name = table + " as t"
    keys, update_param_list = make_keys_multicol(columns_to_update_lst)
    from_clause = get_from_clause_multicol(query_values_dict_lst, columns_to_query_lst, columns_to_update_lst)
    as_clause = get_as_clause_multicol(columns_to_query_lst, update_param_list)
    where_clause = get_where_clause(columns_to_query_lst)
    statement = " ".join([_prefix, table_name, "SET", keys, from_clause, as_clause, where_clause])
    values = get_values_multicol(columns_to_query_lst, columns_to_update_lst, query_values_dict_lst)

    if print_debug_log:
        logging.info("Updating multiple rows into db: %s, %s" % (statement, values))
    return statement, values




from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand
import codecs
import os
import sys
import re

here = os.path.abspath(os.path.dirname(__file__))

def read(*parts):
    # intentionally *not* adding an encoding option to open
    return codecs.open(os.path.join(here, *parts), 'r').read()

def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")

long_description = read('README.rst')

class PyTest(TestCommand):
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = ['--strict', '--verbose', '--tb=long', 'tests']
        self.test_suite = True

setup(
    name='pg_python',
    version=find_version('pg_python', '__init__.py'),
    url='http://github.com/ingloriousb/pg_python',
    license='MIT License',
    author='Gaurav Bhati',
    install_requires=['xlrd',
                      'beautifulsoup4',
                      'psycopg2',
                      'scrapy',
                      'selenium'
                      ],
    author_email='gaurav.hawker@gmail.com',
    description='A Library to communicate with postgres from python',
    #long_description=long_description,
    packages=['pg_python'],
    include_package_data=True,
    platforms='any',
    zip_safe=False,
    classifiers = [
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Development Status :: 4 - Beta',
        'Natural Language :: English',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        ]
)



from urllib.parse import urlparse

def get_domain(url):
    parsed_uri = urlparse(url)
    domain = ('{uri.netloc}/'.format(uri=parsed_uri)).strip("/")
    return domain

def get_proper_domain(url):
    d = get_domain(url)
    prefix_remove = ["http://", "https://", "www."]
    for prefix in prefix_remove:
        if d.startswith(prefix):
            d = d[len(prefix):]
    return d


import requests
from scrapy.http.response.html import HtmlResponse
from scrapy.http.response.xml import XmlResponse
import logging

"""
Response = HTMLReponse of requests.get
"""
MAX_NUM_RETRY = 3


def get_response(url, headers=None, cookies=None, delay=30, response_type="html"):
    num_retries = 0
    response = None
    if cookies is None:
        cookies = {}
    while num_retries < MAX_NUM_RETRY:
        try:
            response = None
            if headers is not None:
                response = requests.get(url, headers=headers,timeout=delay, verify=False, cookies=cookies)
            else:
                response = requests.get(url, timeout=delay, verify=False, cookies=cookies)
            num_retries += 1
            if response.status_code >= 200:
                if response_type == "html":
                  ret_obj = HtmlResponse(url, status=response.status_code, body=response.content, encoding='utf-8')
                  return ret_obj
                elif response_type == "xml":
                  ret_obj = XmlResponse(url, status=response.status_code, body=response.content, encoding='utf-8')
                  return ret_obj
                else:
                    raise Exception("Invalid response type")

        except Exception as e:
            logging.error("Exception %s" %e.message)
            num_retries += 1
    logging.error("Could not fetch the url")
    if response_type == "html":
        err_obj = HtmlResponse(url, status=110, body="<html><body>Failure</body></html>", encoding='utf-8')
    else:
        err_obj = XmlResponse(url, status=110, body="<html><body>Failure</body></html>", encoding='utf-8')
    return err_obj




if __name__ == "__main__":
    for i in range(0, 100):
        r = get_response('https://eprocure.gov.in/cppp/latestactivetenders', delay=60,response_type="html")
        logging.info(r.status)





from ._db_object import Db
from ._write import make_postgres_write_statement, make_postgres_write_multiple_statement
from ._read import make_postgres_read_statement, prepare_values
from ._update import make_postgres_update_statement
from ._update import make_postgres_update_multiple_statement
from ._delete import make_postgres_delete_statement
from ._update import make_postgres_update_multiple_column_statement
from ._update import check_parameters_multicol
import logging
import signal

db_dict = {}


class timeout:
    """Need because of sometimes query will take very long
    (cause of connection already killed and tring to execute query)"""
    def __init__(self, seconds=1, error_message='Timeout'):
        self.seconds = seconds
        self.error_message = error_message

    def handle_timeout(self, signum, frame):
        raise TimeoutError(self.error_message)

    def __enter__(self):
        signal.signal(signal.SIGALRM, self.handle_timeout)
        signal.alarm(self.seconds)

    def __exit__(self, type, value, traceback):
        signal.alarm(0)


def get_db(server="default"):
    db_obj = db_dict.get(server, None)
    return db_obj

print_debug_log = True


def server_connection_check(func):
    def wrapper(*args, **kwargs):
        server = kwargs.get('server', 'default')
        db_obj = get_db(server)
        if db_obj is None:
            logging.info('could not got db object for %s' % server)
            logging.info('db dict %s' % str(db_dict))
        if db_obj.connection.closed != 0:
            logging.info('reconnection to db because of connection closed %s' % db_obj.connection.closed)
            db_obj._make_connection()
        else:
            cursor = db_obj.get_cursor()
            try:
                with timeout(seconds=10):
                    cursor.execute('SELECT 1', [])
            except TimeoutError as e:
                logging.info('timeout occurred, reconnecting dbs')
                db_obj._make_connection()
            except Exception as e:
                logging.info('reconnection to db because of %s' % e)
                db_obj._make_connection()
        return func(*args, **kwargs)
    return wrapper


def pg_server(db_name, username, password, host_address, debug=True, server="default", send_keep_alive_probes=False, socket_idle_time=120, application_name='pg_python'):
    global print_debug_log
    # no need to make socket so, send_keep_alive_probes will no longer usefull
    params_map = {
        'dbname'  : db_name,
        'user'    : username,
        'password': password,
        'host'    : host_address,
        'application_name': application_name
    }
    db_obj = Db(params_map)
    db_dict[server] = db_obj
    print_debug_log = debug
    logging.info('connected to %s with server %s' % (host_address, server))
    return db_obj


@server_connection_check
def write(table, kv_map, server="default"):
    """
    :param table: String.
    :param kv_map: Key values.
    :param server: Alias of the server
    :return success_bool:
    """
    global print_debug_log
    db_obj = get_db(server)
    connection = db_obj.get_connection()

    command, values = make_postgres_write_statement(table, kv_map, print_debug_log)
    try:
        cursor = db_obj.get_cursor()
        cursor.execute(command, values)
        connection.commit()

    except Exception as e:
        logging.error("Db Cursor Write Error: %s" % e)
        db_obj_reconnect = Db(db_obj.params)
        db_dict.pop(server)
        db_dict[server] = db_obj_reconnect
        return False
    return True


@server_connection_check
def read(table, keys_to_get, kv_map, limit=None, order_by=None, order_type=None,
         clause="=", group_by=None, join_clause=' AND ', server="default",
         cols_keep_raw_type=[]):
    """

    :param table: String
    :param keys_to_get: list of strings
    :param kv_map: key value map, if this is None, then limit is maxed at 1000
    :param limit: None or integer
    :param order_by: None or must be of a type String
    :param order_type: String None, "ASC" or "DESC" only

    :param clause: a clause other than default " = " clause between where key value pairs.
            e.g. "where x in y" implies clause will be "in":
             Currently we support a single clause in one function call.

    :param group_by: sql group by
    :param join_clause: the operation between different where clauses
    :param server: server types
    :param cols_keep_raw_type: cols where we want to keep the raw data type.
    :return: values in an array of key value maps
    """
    error_return = None
    db_obj = get_db(server)
    cursor = db_obj.get_cursor()
    command, values = make_postgres_read_statement(table, kv_map, keys_to_get,
                                                   limit, order_by, order_type, print_debug_log,
                                                   clause, group_by, join_clause)

    try:
        cursor.execute(command, values)
        all_values = cursor.fetchall()
        return prepare_values(all_values, keys_to_get, cols_keep_raw_type=cols_keep_raw_type)
    except Exception as e:
        logging.warning("Db Cursor Read Error: %s" % e)
        return []


@server_connection_check
def update(table, update_kv_map, where_kv_map, clause='=', server="default"):
    """
    :param table: table name, type string
    :param update_kv_map: the NEW keyvalue map for values to be updated
    :param where_kv_map: the kv map to search for values, all values ARE ANDed.
    :param clause:
    :param server:
    :return: Success or Failure.
    """
    global print_debug_log
    db_obj = get_db(server)
    connection = db_obj.get_connection()
    cursor = db_obj.get_cursor()
    command, values = make_postgres_update_statement(table, update_kv_map, where_kv_map,
                                                     clause, print_debug_log)
    try:
        cursor.execute(command, values)
        return_dict = {'Status': True}
        if print_debug_log:
            logging.info("%s Record(s) Updated", cursor.rowcount)
        connection.commit()
    except Exception as e:
        logging.error("Db Cursor Update Error: %s" % e)
        db_obj_reconnect = Db(db_obj.params)
        db_dict.pop(server)
        db_dict[server] = db_obj_reconnect
        return False
    return return_dict


@server_connection_check
def read_raw(command, values, server="default"):
    """
    :param table: String
    :param keys_to_get: list of strings
    :param kv_map: key value map, if this is None, then limit is maxed at 1000
    :param limit: None or integer
    :param order_by: None or must be of a type String
    :param order_type: String None, "ASC" or "DESC" only
    :return: values in an array of key value maps
    """
    db_obj = get_db(server)
    cursor = db_obj.get_cursor()
    try:
        if values not in [None, [], {}]:
            cursor.execute(command, values)
        else:
            cursor.execute(command)
        all_values = cursor.fetchall()
        return all_values
    except Exception as e:
        logging.warning("Db Cursor Read Error: %s" % e)
        return []


@server_connection_check
def write_raw(command, values, server="default"):
    """
    :params command, values. Execution commands dirctly for postgres
    """
    global print_debug_log
    db_obj = get_db(server)
    connection = db_obj.get_connection()
    cursor = db_obj.get_cursor()
    try:
        cursor.execute(command, values)
        connection.commit()
    except Exception as e:
        logging.warning("Db Cursor Write Error: %s" % e)
        db_obj_reconnect = Db(db_obj.params)
        db_dict.pop(server)
        db_dict[server] = db_obj_reconnect
        return False
    return True


@server_connection_check
def update_raw(command, server="default"):
    """
    Update statement in the raw format,
    :param command: SQL command
    :return: number of rows affected
    """
    global print_debug_log
    db_obj = get_db(server)
    connection = db_obj.get_connection()
    cursor = db_obj.get_cursor()
    try:
        cursor.execute(command)
        rowcount = cursor.rowcount
        connection.commit()
    except Exception as e:
        logging.warning("Db Cursor Update Error: %s" % e)
        db_obj_reconnect = Db(db_obj.params)
        db_dict.pop(server)
        db_dict[server] = db_obj_reconnect
        return -1
    return rowcount


class bcolors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'


def print_ok(s):
    print((bcolors.OKGREEN + s + bcolors.ENDC))


def print_warn(s):
    print((bcolors.WARNING + s + bcolors.ENDC))


def print_fail(s):
    print((bcolors.FAIL + s + bcolors.ENDC))


def close(server="default"):
    logging.warning("Closing connection for server %s" %server)
    db_obj = get_db(server)
    connection = db_obj.get_connection()
    try:
        cursor = db_obj.get_cursor()
        db_obj.close_cursor(cursor)
        db_obj.close_connection()
    except Exception as e:
        logging.error("Could not close connection properly: %s" % e)


def close_all():
    for server in list(db_dict.keys()):
        close(server)


@server_connection_check
def delete(table, where_kv_map, server="default"):
    """
    Delete the rows resulting from the mentined kv map. No limit.
    :param table: table name, must be string
    :param where_kv_map: the kv map to search for values, all values ARE ANDed.
    :return: True or False
    """
    global print_debug_log
    db_obj = get_db(server)
    connection = db_obj.get_connection()
    cursor = db_obj.get_cursor()
    command, values = make_postgres_delete_statement(table, where_kv_map, print_debug_log)
    try:
        cursor.execute(command, values)
        logging.warning("%s Record(s) Deleted" %cursor.rowcount)
        connection.commit()
    except Exception as e:
        logging.error("Db Cursor Delete Error: %s" % e)
        db_obj_reconnect = Db(db_obj.params)
        db_dict.pop(server)
        db_dict[server] = db_obj_reconnect
        return False
    return True


def check_parameters(column_to_update, columns_to_query_lst, query_values_dict_lst):
    """
    check_prarameters checks whether the passed parameters are valid or not.
    :param column_to_update: name of column that is to be updated.
    :param columns_to_query_lst: list of column names that is used in where clause.
    :param query_values_dict_lst: list of dictionaries containing values for where clause and target column.
    :return: boolean
    """
    # check if dimensions are correct.
    expected_length = 1 + len(columns_to_query_lst)
    all_columns_name = ["update"] + columns_to_query_lst
    flag = 0
    for dict_val in query_values_dict_lst:
        # check dimensions.
        if len(dict_val) != expected_length:
            logging.error("%s doesn't match the dimensions" % (dict_val))
            return False

        # check columns present.
        for column in all_columns_name:
            if column not in dict_val:
                logging.error("%s column isn't present in dictionary" % (column))
                return False
    return True


@server_connection_check
def update_multiple(table, column_to_update, columns_to_query_lst,
                    query_values_dict_lst, server="default", typecast = ""):
    """
    Multiple update support in pg_python
    :param table: table to update into
    :param column_to_update: Single column for set clause
    :param columns_to_query_lst: column names for where clause
    :param query_values_dict_lst: values for where and Set.
    :param server: server from the pg_server's db_obj list
    :param typecast: typecase statement to be appended after c.update e.g. "::int", must be a string
    :return:
    """
    global print_debug_log
    db_obj = get_db(server)
    connection = db_obj.get_connection()
    cursor = db_obj.get_cursor()
    is_parameters_correct = check_parameters(column_to_update, columns_to_query_lst, query_values_dict_lst)
    if not is_parameters_correct:
        logging.error("ERROR in parameters passed")
        return
    if not isinstance(typecast, str):
        logging.error("typecase param is not a string, it must be a string")
    command, values = make_postgres_update_multiple_statement(table,
                                                              column_to_update,
                                                              columns_to_query_lst,
                                                              query_values_dict_lst,
                                                              print_debug_log,
                                                              typecast_suffix=typecast)
    try:
        cursor.execute(command, values)
        return_dict = {'Status': True}
        count = cursor.rowcount
        return_dict['rowcount'] = count

        connection.commit()
    except Exception as e:
        logging.warning("Db Cursor update_multiple Error: %s" % e)
        db_obj_reconnect = Db(db_obj.params)
        db_dict.pop(server)
        db_dict[server] = db_obj_reconnect
        return {'status': False}
    return return_dict


@server_connection_check
def update_multiple_col(table, columns_to_update_lst, columns_to_query_lst, query_values_dict_lst, server="default"):
    """
    Multiple update support in pg_python
    :param table: table to update into, string
    :param columns_to_update_lst: list of column names to be updated for set clause, list of strings
    :param columns_to_query_lst: list of column names to be queried for where clause, list of strings
    :param query_values_dict_lst: list of dictionaries with values for where query and update params. Has 2 items-
            update dictionary - with key='update' and value as dictionary of column name and values to be updated
            where dictionary - with key='where' and value as dictionary of column names and values to be queried
            e.g. [{'where':{'id':1}, 'update':{'col1':'ABC', 'col2':'XYZ'}}]
    :return: dictionary of status (boolean) and num of updated records - {'status':True,'updated_records':1}
    """
    global print_debug_log
    db_obj = get_db(server)
    connection = db_obj.get_connection()
    cursor = db_obj.get_cursor()
    is_parameters_correct = check_parameters_multicol(columns_to_update_lst, columns_to_query_lst,
                                                      query_values_dict_lst)
    if not is_parameters_correct:
        logging.error("ERROR in parameters passed")
        return {'status': False}

    command, values =  make_postgres_update_multiple_column_statement(table,
                                                                      columns_to_update_lst,
                                                                      columns_to_query_lst,
                                                                      query_values_dict_lst,
                                                                      print_debug_log)
    try:
        cursor.execute(command, values)
        count = cursor.rowcount
        connection.commit()
    except Exception as e:
        logging.warning("Db Cursor update_multiple Error: %s" % e)
        db_obj_reconnect = Db(db_obj.params)
        db_dict.pop(server)
        db_dict[server] = db_obj_reconnect
        return {'status': False}
    return {'status':True, 'updated_records':count}


def check_multiple_insert_param(columns_to_insert, insert_values_dict_lst):
    """
    Checks if the pararmeter passed are of correct order.
    :param columns_to_insert:
    :param insert_values_dict_lst:
    :return:
    """
    column_len = len(columns_to_insert)
    for row in insert_values_dict_lst:
        if column_len != len(row):
            logging.error("%s doesn't match the dimensions" % (row))
            return False
        for column in columns_to_insert:
            if column not in row:
                logging.error("%s column isn't present in dictionary" % (column))
                return False
    return True


@server_connection_check
def insert_multiple(table, columns_to_insert_lst, insert_values_dict_lst, server="default"):
    """
    Multiple row insert in pg_python
    :param table: table to insert into.
    :param columns_to_insert_lst: columns value provided.
    :param insert_values_dict_lst: values of corresponding columns.
    :return:
    """
    global print_debug_log
    db_obj = get_db(server)
    connection = db_obj.get_connection()
    cursor = db_obj.get_cursor()
    is_pararmeters_correct = check_multiple_insert_param(columns_to_insert_lst, insert_values_dict_lst)
    if not is_pararmeters_correct:
        logging.error("ERROR in parameters passed")
        return
    command, values = make_postgres_write_multiple_statement(table, columns_to_insert_lst, insert_values_dict_lst,
                                                             print_debug_log)
    try:
        cursor.execute(command, values)
        connection.commit()
    except Exception as e:
        logging.error("Db Cursor Write Error: %s" % e)
        db_obj_reconnect = Db(db_obj.params)
        db_dict.pop(server)
        db_dict[server] = db_obj_reconnect
        return False
    return True
